#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# scores.py

Module (neuralnetwork): Structures to score and keep scores for evaluated data.

"""

from __future__ import division

import logging
import pickle
from dataclasses import dataclass
from typing import Optional, Iterable, Any

import numpy as np
import pandas as pd
from sklearn.metrics import classification_report

from graphs.plot import Plot

logger = logging.getLogger(__name__)


def save(save_file_path, team):
    with open(save_file_path, "wb") as f:
        pickle.dump(team, f)


def load(save_file_path):
    with open(save_file_path, "rb") as f:
        return pickle.load(f)


@dataclass
class Scoring(object):
    def __init__(
        self,
        pred: np.ndarray,
        true: np.ndarray,
        desc: Optional[str] = "",
        mat: bool = False,
    ) -> None:
        """
        Class to manage scoring variables from fitted classifiers. 

        Parameters
        ----------
        pred : ndarray
            Fitted model's "predicted" output.
        true : ndarray
            Descriptors of each predictable value.
        mat : bool, optional
            Whether to output a Confusion Matrix. The default is False.

        Returns
        -------
        None.
        """
        # Input variables
        self.predicted: Iterable[Any] = pred
        self.true: Iterable[Any] = true
        self.classes: list = list(np.unique(self.predicted))
        self.descriptor: str = desc
        self.report: pd.DataFrame = self.get_report()

        if mat:
            self.mat: Any = self.get_confusion_matrix()
        if desc is None:
            logging.info("No descriptor")
            pass

    def get_report(self) -> pd.DataFrame:
        """ Get classification report"""
        if self.descriptor:
            assert self.descriptor in [
                "train",
                "training",
                "test",
                "testing",
                "eval",
                "val",
                "evaluate",
            ]

        self.report = classification_report(
            self.true,
            self.predicted,
            target_names=self.classes,
            labels=self.classes,
            output_dict=True,
        )
        report_df = pd.DataFrame(data=self.report).transpose()
        return report_df

    def get_confusion_matrix(self, caption: Optional[str] = "") -> object:
        """ Get confusion matrix"""
        mat = Plot.confusion_matrix(
            y_true=self.true,
            y_pred=self.predicted,
            labels=self.classes,
            caption=caption,
        )

        return mat



"""
#file_helpers.py

Module(utils/file_helpers): File handling data-container class to keep all file-related
data.
"""
from __future__ import annotations

import logging
from collections import namedtuple
from pathlib import Path
from typing import Optional

import pandas as pd

from canalysis.helpers import funcs

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG, format="%(message)s")
logger.setLevel(logging.DEBUG)


class FileHandler:
    """
    File handler. From directory, return genorator looping through files.

    Parameters:
    ___________
    animal: str
        - Animal ID, letter + number combo
    date: str
        - Session date, all_numeric
    _directory: str
        - Path as str, contains /
    _tracename: Optional[str]
        - Pattern matching to apply for trace files
    _eventname: Optional[str]
        - Pattern matching to apply for event files

    Directory structure:
    ___________
    
    ./directory
        -| Animal
        --| Date
        ---| Results
        -----| Graphs
        -----| Statistics
        ---| Data_traces*
        ---| Events_gpio (or)
        ---| Events_gpio_processed*
        Use tree() to print current directory tree.
    """

    def __init__(
        self,
        animal,
        date,
        dir,
        tracename: Optional[str] = "traces",
        eventname: Optional[str] = "processed",
        _gpioname: Optional[str] = "gpio.csv",
        eatingname: Optional[str] = None,
    ) -> None:

        # TODO: glob for .csv and .xlxs
        self.animal = animal
        self.date = date
        self._directory: Path = Path(dir)
        self._tracename: Optional[str] = tracename
        self._eventname: Optional[str] = eventname
        self._gpioname: Optional[str] = _gpioname
        self._eatingname: Optional[str] = eatingname

        self.color_dict: namedtuple
        self._validate()

        self.session: str = f"{animal}_{date}"
        self.animaldir: Path = Path(self._directory / self.animal)
        self.sessiondir: Path = Path(self.animaldir / self.date)
        if not self.sessiondir.is_dir():
            raise NotADirectoryError(f"Directory '{self.sessiondir}' is not a valid directory or it does not exist, "
                                     f"check params.yaml Directory entry.")
        self._gpio_file: Optional[bool] = False
        self._make_dirs()

    def _validate(self):
        """Validate format of input data. Raises AttributeErrors for each check."""
        if not funcs.check_numeric(self.date):
            raise AttributeError(f"Date must be all numeric, not {self.date}")
        if not funcs.check_path(self._directory):
            raise AttributeError(f'Directory must contain "/", not {self._directory}')
        if funcs.check_numeric(self.animal) or funcs.check_path(self.animal):
            raise AttributeError(
                f"Animal must not be only numeric or contain path characters, "
                f"{self.animal}"
            )

    @property
    def directory(self) -> Path:
        return self._directory

    @directory.setter
    def directory(self, new_dir: str) -> None:
        self._directory: str = new_dir

    @directory.setter
    def directory(self, new_dir: str) -> None:
        self._directory: str = new_dir

    @property
    def tracename(self) -> str:
        return self._tracename

    @tracename.setter
    def tracename(self, new_tracename: str) -> None:
        self._tracename: str = new_tracename

    @property
    def eventname(self) -> str:
        return self._eventname

    @eventname.setter
    def eventname(self, new_eventname: str) -> None:
        self._eventname: str = new_eventname

    @property
    def gpioname(self) -> str:
        return self._gpioname

    @gpioname.setter
    def gpioname(self, new_gpioname: str):
        self._gpioname: str = new_gpioname

    @property
    def eatingname(self) -> str:
        return self._eatingname

    @eatingname.setter
    def eatingname(self, new_eatingname: str):
        self._eatingname: str = new_eatingname

    def get_traces(self) -> list[Path]:
        for x in self.sessiondir.iterdir():
            print(x)
        return [p for p in self.sessiondir.glob(f"*{self._tracename}*.csv")]

    def get_events(self) -> list[Path]:
        return [p for p in self.sessiondir.glob(f"*{self._eventname}*")]

    def get_gpio_files(self) -> list[Path]:
        return [p for p in self.sessiondir.glob(f"*{self._gpioname}*")]

    def get_eating_files(self) -> list[Path]:
        return [p for p in self.sessiondir.glob(f"*{self._eatingname}*")]

    def get_tracedata(self) -> pd.DataFrame:
        tracefiles: list[Path] = self.get_traces()
        if not tracefiles:
            files = self.search_files()
            raise FileNotFoundError(
                f'No files in {self.sessiondir} matching "{self._tracename}"'
                f"Files found: {files}"
            )

        if len(tracefiles) > 1:
            logging.info(
                f'Multiple trace-files found in {self.sessiondir} matching "'
                f'{self._tracename}":'
            )
            for tracefile in tracefiles:
                logging.info(f"{tracefile.stem}")
        logging.info(f"Taking file: {tracefiles[0].stem}")
        return pd.read_csv(str(tracefiles[0]), low_memory=False)

    def get_eventdata(self) -> pd.DataFrame:
        eventfiles: list[Path] = self.get_events()
        if eventfiles is None:
            raise FileNotFoundError(
                f'No files in {self.sessiondir} matching "{self._eventname}"'
            )
        if len(eventfiles) > 1:
            logging.info(
                f'Multiple event-files found in {self.sessiondir} matching "'
                f'{self._eventname}":'
            )
            for event_file in eventfiles:
                logging.info(f"{event_file}")
            logging.info(f"Taking file: {eventfiles[0]}")
        return pd.read_csv(str(eventfiles[0]), low_memory=False)

    def get_gpiodata(self) -> pd.DataFrame:
        gpiofiles: list[Path] = self.get_gpio_files()
        if gpiofiles is None:
            raise FileNotFoundError(
                f'No files in {self.sessiondir} matching "{self._gpioname}"'
            )
        if len(gpiofiles) > 1:
            self._gpio_file = True
            logging.info(
                f'Multiple gpio-files found in {self.sessiondir} matching "'
                f'{self._gpioname}":'
            )
            for gpio_file in gpiofiles:
                logging.info(f"{gpio_file}")
            logging.info(f"Taking file: {gpiofiles[0]}")
        return pd.read_csv(str(gpiofiles[0]), low_memory=False)

    def get_eatingdata(self) -> pd.DataFrame:
        eatingfiles: list[Path] = self.get_eating_files()
        if eatingfiles is None:
            raise FileNotFoundError(
                f'No files in {self.sessiondir} matching "{self._eatingname}"')
        if len(eatingfiles) > 1:
            logging.info(
                f'Multiple eating-files found in {self.sessiondir} matching "'
                f'{self._eatingname}":')
            for eating_file in eatingfiles:
                logging.info(f"{eating_file}")
            logging.info(f"Taking file: {eatingfiles[0]}")
        logging.info("Eating data set.")
        return pd.read_csv(
            str(eatingfiles[0]),
            low_memory=False,
            header=0,
            usecols=['Marker Name', 'TimeStamp', 'TimeStamp2'])

    def unique_path(self, filename) -> Path:
        counter = 0
        while True:
            counter += 1
            path = self.sessiondir / filename
            if not path.exists():
                return path

    def _make_dirs(self) -> None:
        self.sessiondir.parents[0].mkdir(parents=True, exist_ok=True)
        return None

    def get_cwd(self) -> str:
        return str(self._directory.cwd())

    def get_home_dir(self) -> str:
        return str(self._directory.home())

    def tree(self) -> None:
        print(f"-|{self._directory}")
        for path in sorted(self.sessiondir.rglob("[!.]*")):  # exclude .files
            depth = len(path.relative_to(self.sessiondir).parts)
            spacer = "    " * depth
            print(f"{spacer}-|{path.name}")
            return None

    def search_files(self):
        dirpath = self._directory
        assert dirpath.is_dir()
        file_list = []
        for x in dirpath.iterdir():
            if x.is_file():
                file_list.append(x)
        return file_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# eventdata.py
Module (data.data_utils): Process event/gpio data exported from _video_gpio.isxd file.
"""
from __future__ import annotations

from typing import Sized, Iterable, Any
import logging
from dataclasses import dataclass, field
import numpy as np
import pandas as pd
from numpy import ndarray

from canalysis.data.data_utils.file_handler import FileHandler
from canalysis.helpers import funcs


@dataclass(order=False)
class EventData:
    filehandler: FileHandler
    color_dict: dict
    tracedata_time: np.ndarray
    timestamps: field = field(init=False, default_factory=dict)
    trial_times: field = field(init=False, default_factory=dict)
    # Initialize empty placeholders to fill later
    numlicks: Sized | int = field(default_factory=list)
    drylicks: list = field(default_factory=list)
    __allstim: list = field(default_factory=list)
    alltastestim: list = field(default_factory=list)
    nonreinforced: Iterable = field(default_factory=list)
    matched: bool = False

    def __post_init__(self, ):
        self.timestamps: dict = self.__get_timestamps()
        self.drylicks = [x for x in self.timestamps["Lick"] if x not in self.__allstim]
        self.trial_times: dict = self.__get_trial_times()
        self.nonreinforced: ndarray = self.__get_nonreinforced()

    def __repr__(self):
        return type(self).__name__

    def __len__(self, ):
        return len(self.numlicks)

    def __get_timestamps(self, ):
        timestamps: dict = {}
        data: pd.DataFrame = self.filehandler.get_eventdata()
        events = data.rename(columns={"Time(s)": "time"})
        events['time'] = funcs.get_matched_time(self.tracedata_time, events['time'])
        for stimulus in events.columns[1:]:
            timestamps[stimulus] = list(
                events["time"].iloc[np.where(events[stimulus] == 1)[0]]
            )
            if stimulus != "Lick":
                self.__allstim.extend(timestamps[stimulus])
            if stimulus != "Lick" and stimulus != "ArtSal":
                self.alltastestim.extend(timestamps[stimulus])
        self.numlicks: Sized | int = len(timestamps["Lick"])
        self.alltastestim.sort()
        return timestamps

    def __get_nonreinforced(self, ):
        times = []
        lickstamps = np.array(self.timestamps["Lick"])
        intervals = funcs.interval(self.alltastestim, 2)
        for iteration, interv in enumerate(intervals):
            times.extend(
                lickstamps[
                    np.where((lickstamps >= interv[0]) & (lickstamps <= interv[1]))
                ]
            )
        nr = np.setdiff1d(lickstamps, times)
        return nr

    def __get_trial_times(self) -> dict:
        trial_times: dict = {}
        for stim, tslist in self.timestamps.items():
            if stim != "Lick" and stim != "ArtSal" and len(self.timestamps[stim]) > 0:
                # Get list of tastant deliveries
                tslist.sort()
                tslist = np.array(tslist)
                times = [tslist[0]]  # First one is a trial
                # Delete duplicates
                for i in range(0, len(tslist) - 1):
                    if tslist[i] == tslist[i + 1]:
                        nxt = tslist[i + 1]
                        tslist = np.delete(tslist, i + 1)
                        print(f"Deleted timestamp {nxt}")
                # Add each tastant delivery preceded by a drylick (trial start).
                for ts in tslist:
                    last_stimtime = tslist[np.where(tslist == ts)[0] - 1]
                    last_drytime = self.drylicks[np.where(self.drylicks < ts)[0][-1]]
                    if last_drytime > last_stimtime:
                        times.append(ts)
                trial_times[stim] = times
        return trial_times

    def get_trials(self) -> None:
        """Print number of instances of each event"""
        for stim, trials in self.trial_times.items():
            logging.info(f"{stim} - {len(trials)}")
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 12 09:17:47 2022

@author: flynnoconnell
"""


class DataFrameError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "DataFameError has been raised"


class ComponentError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "Number components was chosen to be larger than the total number of cells (features)."


class DuplicateError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "A duplicate time may have been chosen for time of peak"


class ParameterError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "Only rbf or linear kernals allowed for this analysis."


class MergeError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "Cant merge these indices."


class FileError(Exception):
    def __init__(self, msg: str, filetype: str = "Traces"):
        if not msg:
            raise AttributeError("Message param missing from FileError call")
        else:
            self.message = msg
            self.filetype = filetype

    def __str__(self):
        return f"{self.message} - File type: {self.filetype}"


class MatchError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "Time to match (match) larger than iterable to match to, likely wrong order of match/time."


class PCAError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "{0} ".format(self.message)
        else:
            return "DataFrame not properly set, missing column for 'colors' or 'events' .. "



from matplotlib.figure import Figure
import matplotlib.pyplot as plt


class CalFigure(Figure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tight_layout()
        self.set_dpi(300)

    def close(self):
        plt.close(self)

    def save(self, dir):
        plt.savefig(dir)










#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# .../neuralnetwork/utils/funcs.py

Functions for neural network classes.
"""

import pathlib

import numpy as np


def create_experiment_log_dir(root: str, parents: bool = True) -> str:
    root_path = pathlib.Path(root).resolve()
    child = (
        create_from_missing(root_path)
        if not root_path.exists()
        else create_from_existing(root_path)
    )
    child.mkdir(parents=parents)
    return child.as_posix()


def create_from_missing(root: pathlib.Path) -> pathlib.Path:
    return root / "0"


def create_from_existing(root: pathlib.Path) -> pathlib.Path:
    children = [
        int(c.name) for c in root.glob("*") if (c.is_dir() and c.name.isnumeric())
    ]
    if is_first_experiment(children):
        child = create_from_missing(root)
    else:
        child = root / increment_experiment_number(children)
    return child


def is_first_experiment(children: list[int]) -> bool:
    return len(children) == 0


def increment_experiment_number(children: list[int]) -> str:
    return str(max(children) + 1)


def shuffle_data(X, y, seed=None):
    """ Random shuffle of the samples in X and y """
    if seed:
        np.random.seed(seed)
    idx = np.arange(X.shape[0])
    np.random.shuffle(idx)
    return X[idx], y[idx]




from .all_data import AllData
from .eating_data import EatingData
from .event_data import EventData
from .gpio_data import GpioData
from .taste_data import TasteData
from .trace_data import TraceData
from .calcium_data import CalciumData

__all__ = ["TasteData", "EatingData", "AllData", "GpioData", "EventData", "TraceData", "CalciumData"]



# -*- coding: utf-8 -*-
"""
# process_gpio.py

"""
from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import pandas as pd

from canalysis.data.data_utils.file_handler import FileHandler


@dataclass
class GpioData:
    filehandler: FileHandler = FileHandler
    threshold: int = 3000
    dist_adjust: float = 0.005
    decode_gpio = {
        "Rinse": [1, 2, 3, 4],
        "ArtSal": [1, 3, 4],
        "Quinine": [1, 2, 4],
        "NaCl": [1, 3],
        "Sucrose": [1, 2],
        "MSG": [1, 4],
        "Citric": [1, 2, 3],
    }
    gpiodata = None

    def __post_init__(self):
        self.gpiodata = self.filehandler.get_gpiodata()
        self.timestamps = {}

    @staticmethod
    def within(arr, ts, adjust):
        arr = np.array(arr)
        out = arr[np.where((arr > ts - adjust) & (arr < ts + adjust))]
        return out

    def get_rec_end(self):
        rec_end = max(
            self.gpiodata.loc[
                np.where(self.gpiodata[" Channel Name"] == " BNC Sync Output")[0],
                "Time (s)",
            ]
        )
        return rec_end

    def trim(self):
        self.gpiodata = self.gpiodata.iloc[
            np.where(self.gpiodata[" Channel Name"].str.contains("GPIO"))[0], :
        ]

    def get_timestamps(self):
        for chan in pd.unique(self.gpiodata[" Channel Name"]):
            print("Processing events for:", chan)
            event = []
            gp_chan = self.gpiodata.iloc[
                np.where(self.gpiodata[" Channel Name"] == chan)[0], :
            ]
            gp_check = gp_chan[gp_chan[" Value"] > self.threshold]
            for index, row in gp_check.iterrows():
                t = row["Time (s)"]
                # see if any times in the last 11 ms exceed the threshold.
                # pulses last 10 ms, should not any true values
                tcomp = (
                    gp_chan[
                        (gp_chan["Time (s)"] >= t - 0.011) & (gp_chan["Time (s)"] < t)
                    ][" Value"]
                    > self.threshold
                )
                # if all checks are good add the event
                if row[" Value"] > self.threshold and not np.any(tcomp):
                    event.append(t)
            self.timestamps[chan] = event
        self.timestamps["Lick"] = self.timestamps[" GPIO-1"]

    def _clean_validate(self):
        for i in range(2, 5):  # for channels 2 through 4
            for ts in self.timestamps[" GPIO-{}".format(i)]:
                # if the timestamp is not in channel 1 (all licks)
                if ts not in self.timestamps[" GPIO-1"]:
                    # get all nearby timestamps from GPIO1
                    tscheck = self.within(
                        self.timestamps[" GPIO-1"], ts, self.dist_adjust
                    )
                    if len(tscheck) == 1:  # there should only be one
                        # if so replace the timestamp with the one from GPIO-1
                        print(
                            "Adjusted GPIO-{} timestamps from {} to {}".format(
                                i, ts, tscheck[0]
                            )
                        )
                        self.timestamps[" GPIO-{}".format(i)].remove(ts)
                        self.timestamps[" GPIO-{}".format(i)].append(tscheck[0])
            self.timestamps[" GPIO-{}".format(i)].sort()

    def _collect_gpio(self):
        allts = []
        # this loop collects all GPIO timestamps
        for chan in pd.unique(self.gpiodata[" Channel Name"]):
            allts.extend(self.timestamps[chan])

    def decode(self):
        for stim, inputs in self.decode_gpio.items():
            ts_holder = []
            for ts in self.timestamps[" GPIO-1"]:  # for each lick
                for chan in range(1, 5):
                    if (
                        ts not in self.timestamps[" GPIO-{}".format(chan)]
                        and chan in inputs
                    ):
                        break  # if the input is needed to code for this stimulus and isn't present, move on
                    elif (
                        ts in self.timestamps[" GPIO-{}".format(chan)]
                        and chan not in inputs
                    ):
                        break  # if the input is not needed to code for this stimulus and is present, move on
                else:
                    ts_holder.append(ts)
            self.timestamps[stim] = ts_holder



from __future__ import annotations

import matplotlib.pyplot as plt
from ..base import _base_figure 
from ..graph_utils import helpers

helpers.update_rcparams()


class BaseHeatmap:
    def __init__(
        self,
        cmap: str = "plasma",
        colorbar: bool = False,
        fig: plt.figure = None,
        ax: plt.axes = None,
        **figargs
    ):
        self.cmap = plt.get_cmap(cmap)
        self.colorbar = colorbar
        self.fig = plt.figure(FigureClass=_base_figure.CalFigure) if fig is None else fig
        self.ax = self.fig.add_subplot(111) if ax is None else ax

    def clear(self):
        self.ax.clear()



from __future__ import annotations

from typing import Optional, Iterable, Any, ClassVar, Sized
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler


def get_pca(data, numcomp: int = 4):
    return _PrincipalComponents(data, numcomp=3)


class _PrincipalComponents:
    def __init__(
            self,
            data: pd.DataFrame,
            numcomp: int
    ):
        self.data: pd.DataFrame = data
        self.numcomp: int = numcomp
        self.pca: ClassVar = None
        self.variance_explained: Iterable[Any] | None = None
        self.pca_df: pd.DataFrame | None = None
        self.fit_pca()

    def __repr__(self):
        return f"{type(self).__name__}, {self.numcomp}"

    def fit_pca(self, ) -> None:
        data_prepped = StandardScaler().fit_transform(self.data)
        self.pca = PCA(n_components=self.numcomp)
        data_fit = self.pca.fit_transform(data_prepped)
        self.variance_explained = np.round(self.pca.explained_variance_ratio_ * 100, decimals=1)
        labels = [
            "PC" + str(x) + f" - {self.variance_explained[x - 1]}%"
            for x in range(1, len(self.variance_explained) + 1)
        ]
        self.pca_df = pd.DataFrame(data_fit, columns=labels)
        return None




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
# neuralnetwork.py

Module (neuralnetwork): Classes/functions for NeuralNetwork module training.
"""

from __future__ import division

import logging
from enum import Enum, auto

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.model_selection import (
    StratifiedShuffleSplit,
    RepeatedStratifiedKFold,
    GridSearchCV,
)
from sklearn.svm import SVC

from graphs.graph_utils.helpers import plot_learning_curve
from neuralnetwork.nn_utils._properties import _validate, _props
from neuralnetwork.nn_utils.datahandler import DataHandler
from neuralnetwork.nn_utils.scores import Scoring

logger = logging.getLogger(__name__)


# %%


class Stage(Enum):
    BASE = auto(int)
    SPLIT = auto(int)
    SCALE = auto(int)


class SupportVectorMachine(_validate, _props):
    def __init__(self, data, target, stage, cv=None):
        """
        Base class for svc.SVM neural network model to handle:
             a binary classification to [-1, 1] values, or 
             a multivariate classification to [-1, n-1]. 
        -Instances of StandardScaler() needed to scale data to a mean = 0 and stdev = 1. 
        -These attributes will be shared for each session.
        
        ..:Steps: 
            1) Split dataset into Train/Train-labels (X_train, Y_train) and Test/Test-labels
            2) Scale data to mean = 0 Stdev = 1,
    and encode labels to integers -1 to 1-n (n = len(labels))
            3) Optimize classifier (optional),
    tune hyperparameters with sklearn.GridSearchCV using default or 
                   custom parameters for C, Gamma, and anything else. 
            4) Fit the model to training data. 
            5) Predict test labels from fitten model.
            6) If scores are sufficient, use fitted model to predict labels from different eval dataset.
                
            Note: Each step can be performed independantly with custom data. 
        
        Parameters
        ----------
        data : pd.DataFrame | np.ndarray
            Features x Samples.
        target : pd.Series | np.ndarray
            Labels for data input.
        stage: str
            Stage of network: train, test, or eval.

        Returns
        -------
        None.

        """
        self.stage = stage
        self.traindata = DataHandler(data=data, target=target)
        self.scaler = preprocessing.StandardScaler()
        self.model = None
        self.grid = None
        # default CV if no other is given
        if not cv:
            self._cv = RepeatedStratifiedKFold()
        else:
            self._cv = cv
        self.trainset = {}
        self.scores_train = None
        self.scores_eval = None

    @staticmethod
    def to_numpy(arg):
        if isinstance(arg, np.ndarray):
            return arg
        elif isinstance(arg, pd.Series):
            return arg.to_numpy()
        else:
            return np.array(arg)

    def split(self, train_size: float = 0.9, n_splits=1, **params) -> None:
        """
        Split training dataset into train/test data.

        Args:
            train_size (float, optional): Proportion used for training. Defaults to 0.9.
            n_splits (int, optional): Number of iterations to split. Defaults to 1.
            **params (dict): Optional ShuffleSplit estimator params or custom labels/data.

        Returns:
            X_train (Iterable[Any]): Training features.
            x_test (Iterable[Any]): Testing features.
            Y_train (Iterable[Any]): Training labels.
            y_test (Iterable[Any]): Testing labels.
        """
        data = params.pop("data", self.traindata.data)
        target = params.pop("target", self.traindata.target)
        if self.cv is None:
            cv = StratifiedShuffleSplit(
                n_splits=n_splits, train_size=train_size, **params
            )
        else:
            cv = self.cv

        train_index, test_index = next(cv.split(data, target))
        X_train, x_test = data[train_index], data[test_index]
        Y_train, y_test = target[train_index], target[test_index]

        self.trainset["X_train"] = X_train
        self.trainset["x_test"] = x_test
        self.trainset["Y_train"] = Y_train
        self.trainset["y_test"] = y_test
        self.scale()

        return None

    def scale(self, **kwargs):
        """
        Scale to mean = 0 and st.dev = 1 a train/split dataset.

        Parameters
        ----------
        **kwargs : dict
            Custom train/test data. Must be supplied given names: 
                -> X_train, x_test, Y_train, y_test
                
        Returns
        -------
        None.
        """
        assert hasattr(self.trainset, "x_test")
        if not kwargs:
            X_train = self.trainset["X_train"]
            x_test = self.trainset["x_test"]
        else:
            X_train = kwargs["X_train"]
            x_test = kwargs["x_test"]
        # Get scaler for only training data, apply to training and test data
        self.scaler.fit(X_train)
        X_train_scaled = self.scaler.transform(X_train)
        x_test_scaled = self.scaler.transform(x_test)
        self.trainset["X_train"] = self.to_numpy(X_train_scaled)
        self.trainset["x_test"] = self.to_numpy(x_test_scaled)
        return None

    def optimize_clf(
        self,
        X_train=None,
        Y_train=None,
        param_grid=None,
        verbose=True,
        refit=True,
        **svcparams,
    ):
        assert hasattr(self.trainset, "x_test")
        if param_grid:
            param_grid = param_grid
        else:
            param_grid = {
                "C": [0.1, 1, 10, 100, 150, 200, 500, 1000],
                "gamma": ["scale", "auto"],
                "kernel": ["linear", "rbf", "poly"],
            }
        if not X_train:
            X_train = self.trainset["X_train"]
            Y_train = self.trainset["Y_train"]
        else:
            X_train = X_train
            Y_train = Y_train

        svc = SVC(class_weight="balanced", **svcparams)
        self.grid = GridSearchCV(
            svc, param_grid=param_grid, cv=self.cv, refit=refit, verbose=verbose
        )
        self.grid.fit(X_train, Y_train.ravel())
        print("**", "-" * 20, "*", "-" * 20, "**")
        print(f"Best params: {self.grid.best_params_}")
        print("**", "-" * 5)
        print(f"Score: {self.grid.best_score_}")
        # Use optimized parameters for SVC model
        kernel_best = self.grid.best_params_["kernel"]
        c_best = self.grid.best_params_["C"]
        gamma_best = self.grid.best_params_["gamma"]
        self.model = SVC(C=c_best, kernel=kernel_best, gamma=gamma_best, verbose=False)
        print("Model optimized")
        return None

    def fit_clf(self, model=None) -> object:
        """
        Fit classifier to training data.

        Args:
            model (model): optional, input custom model.
        """
        assert hasattr(self.trainset, "x_test")
        X_train = self.trainset["X_train"]
        Y_train = self.trainset["Y_train"]
        if model:
            self.model = model  # If model proveed, override current model
            logging.info(f"Model provided has been fit: {model}")
        self.model.fit(X_train, np.ravel(Y_train))
        return None

    def predict_clf(self, x_test=None):
        assert hasattr(self.trainset, "x_test")
        if not x_test:
            x_test = self.trainset["x_test"]
        y_pred = self.model.predict(x_test)
        self.trainset["y_pred"] = y_pred
        y_true = self.trainset["y_test"]
        _validate._full_predict(y_pred, y_true)
        self.scores_train = Scoring(y_pred, y_true, desc="train", mat=True)
        return None

    def evaluate_clf(self):
        y_pred = self.model.predict(self.evaldata.data)
        y_true = self.evaldata.target
        self.trainset["eval_y_pred"] = y_pred
        self.trainset["eval_y_true"] = y_true
        _validate._full_predict(y_pred, y_true)
        self.scores_eval = Scoring(y_pred, y_true, desc="eval", mat=True)
        return None

    def get_learning_curves(self, estimator, title: str = "Learning Curve"):

        import matplotlib.pyplot as plt

        _, axes = plt.subplots(3, 2, figsize=(10, 15))

        X = self.trainset["X_test"]
        y = self.trainset["y_test"]

        # Cross validation with 50 iterations to get smoother mean test and train
        # score curves, each time with 20% data randomly selected as a validation set.
        title = title
        plot_learning_curve(
            estimator, title, X, y, axes=axes[:, 0], ylim=(0, 1.01), cv=self.cv
        )
        kern = self.grid.best_params_["kernel"]
        gam = self.grid.best_params_["gamma"]
        C = self.grid.best_params_["C"]
        title = f"SVC - kernel = {kern}, gamma = {gam}, C = {C}"
        plot_learning_curve(
            estimator, title, X, y, axes=axes[:, 1], ylim=(0, 1.01), cv=self.cv
        )
        plt.show()


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# draw_plots.py

Module (core): Functions for drawing graphs.

"""
from __future__ import annotations

from typing import Optional, Sized

from matplotlib import pyplot as plt
import seaborn as sns
import matplotlib.font_manager as fm
import numpy as np

from canalysis.graphs.graph_utils import helpers, ax_helpers
from canalysis.graphs.base._base_figure import CalFigure
import matplotlib
helpers.update_rcparams()


def get_figure(*args, ax=None, **kwargs):
    fig = plt.figure(*args, FigureClass=CalFigure, **kwargs)
    ax = ax if ax is not None else plt.gca()
    ax.set_facecolor = "w"
    ax.patch.set_facecolor('w')
    return fig, ax


def get_axis_points(data):
    if data.shape[1] > 2:
        return [data.iloc[:, 0], data.iloc[:, 1], data.iloc[:, 2]]
    else:
        return [data.iloc[:, 0], data.iloc[:, 1]]


def pca_scatter(data,
                colors,
                color_dict=None,
                s: int = 10,
                angles: tuple = None):
    matplotlib.use("Qt5Agg")
    proj = '3d' if data.shape[1] > 2 else 'rectilinear'
    ax = plt.axes(projection=proj)
    ax = ax_helpers.get_axis_labels(ax, data)
    ax.set_title(f"{'Principal Components'}", fontweight="bold")
    ax.xaxis.get_label().set_fontsize(10)
    ax.yaxis.get_label().set_fontsize(10)
    ax.zaxis.get_label().set_fontsize(10)
    ax.scatter(
        *get_axis_points(data),
        c=colors,
        s=s,
        marker='o',
        facecolor='w',
    )
    if color_dict is not None:
        ax = ax_helpers.get_legend(ax, color_dict, colors, s)
    fig, ax = get_figure(ax=ax, figsize=(3, 3))
    if angles:
        ax.view_init(*angles)
    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False
    plt.show()
    return fig


def pca_skree(variance: Sized, title: str = "") -> plt.figure:
    """
    Line chart skree plot.

    Parameters
    ----------
    variance : np.ndarray
        From PCA.explained_variance_ratio_.
    title : str, optional
        Title of graph. The default is ''.
    """
    lab = np.arange(len(variance)) + 1
    fig, ax = plt.subplot(111)
    ax.plot(lab, variance, "o-", linewidth=2, color="blue")
    ax.set_xticks(np.arange(0, lab[-1], 1.0))
    ax.set_title(f"{title}" + "Scree Plot")
    ax.set_xlabel("Principal Component")
    ax.set_ylabel("Variance Explained (%)")
    leg = plt.legend(
        ["Eigenvalues from SVD"],
        loc="best",
        borderpad=0.3,
        shadow=False,
        prop=fm.FontProperties(size="small"),
        markerscale=0.4,)
    leg.get_frame().set_alpha(1)
    plt.show()
    return fig


def confusion_matrix(
        y_pred,
        y_true,
        labels: list,
        xaxislabel: Optional[str] = None,
        yaxislabel: Optional[str] = None,
        caption: Optional[str] = "",
) -> np.array:
    """

    Parameters
    ----------
    y_pred : TYPE
        DESCRIPTION.
    y_true : TYPE
        DESCRIPTION.
    labels : list
        DESCRIPTION.
    xaxislabel : Optional[str], optional
        DESCRIPTION. The default is None.
    yaxislabel : Optional[str], optional
        DESCRIPTION. The default is None.
    caption : Optional[str], optional
        DESCRIPTION. The default is ''.

    Returns
    -------
    mat : TYPE
        DESCRIPTION.

    """
    from sklearn.metrics import confusion_matrix
    mat = confusion_matrix(y_pred, y_true)
    sns.heatmap(
        mat.T,
        square=True,
        annot=True,
        fmt="d",
        cbar=False,
        xticklabels=labels,
        yticklabels=labels,
    )
    if xaxislabel:
        plt.xlabel(xaxislabel)
    else:
        plt.xlabel("true label")
    if yaxislabel:
        plt.ylabel(yaxislabel)
    else:
        plt.ylabel("predicted label")
    if caption:
        plt.text(0, -0.03, caption, fontsize="small")
    plt.show()
    return mat



"""Test availability of required packages."""

import unittest
from pathlib import Path

import pkg_resources

_REQUIREMENTS_PATH = Path(__file__).parent.with_name("requirements.txt")


class TestRequirements(unittest.TestCase):
    """Test availability of required packages."""

    def test_requirements(self):
        """Test that each required package is available."""
        # Ref: https://stackoverflow.com/a/45474387/
        requirements = pkg_resources.parse_requirements(_REQUIREMENTS_PATH.open())
        for requirement in requirements:
            requirement = str(requirement)
            with self.subTest(requirement=requirement):
                pkg_resources.require(requirement)


"""
A package for analysis, statistics and visualization of Calcium Imaging data,
specifically from including from Inscopix.
"""
from __future__ import annotations
from inspect import getsourcefile
import os
import logging
import yaml
from canalysis.data.containers import CalciumData
from canalysis.data.data_utils.file_handler import FileHandler

logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)
__location__ = os.path.abspath(getsourcefile(lambda:0))
ROOT_DIR = os.path.dirname(__location__)

class Params:
    """@DynamicAttrs"""
    def __init__(self, parameter_dict):
        for key in parameter_dict:
            setattr(self, key, parameter_dict[key])


def get_parameters():
    with open(os.path.join(ROOT_DIR, 'params.yaml'),
              'rb') as f:
        parameters = yaml.safe_load(f.read())
    return Params(parameters)

def get_data(doeating: bool = True, doevents: bool = True, adjust=None):
    params = get_parameters()
    filehandler = FileHandler(
            params.Session['animal'],
            params.Session['date'],
            params.Directory['data'],
            params.Filenames['traces'],
            params.Filenames['events'],
            params.Filenames['gpio'],
            params.Filenames['eating'],
    )
    return CalciumData(filehandler,
                       doeating=params.Filenames['doeating'],
                       doevents=params.Filenames['doevents'],
                       color_dict=params.Colors,
                       adjust=params.Filenames['adjust'],
                       )
    
# Path: CalciumAnalysis/CalciumAnalysis/data/containers/calcium_data.py
__all__ = [
    "CalciumData",
    "FileHandler",
    "get_data",
    "Params",
    "Plot",
    "Heatmaps",
    "ProcessData",
    ]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# eating_data.py

Module: Classes for food-related data processing.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
import pandas as pd
import numpy as np
from typing import Optional, Generator, Iterable, Any
from canalysis.helpers import funcs
from canalysis.data.data_utils.file_handler import FileHandler
from canalysis.data.containers.trace_data import TraceData
from canalysis.graphs.heatmaps import EatingHeatmap

logger = logging.getLogger(__name__)


@dataclass
class EatingData:
    __filehandler: FileHandler
    __tracedata: TraceData
    color_dict: dict
    adjust: int | float
    eatingdata: pd.DataFrame = field(init=False)
    signals: pd.DataFrame = field(init=False)

    def __post_init__(self,):
        self.raw_eatingdata = self.__filehandler.get_eatingdata().sort_values("TimeStamp")
        # Core attributes
        self.__set_adjust()
        self.__clean()
        self.__match()
        self.eatingdata: pd.DataFrame = self.__set_eating_signals()
        self.signals = self.eatingdata.drop(columns=['event', 'color'])
        self.events: pd.Series = self.eatingdata['event']
        self.colors: pd.Series = self.eatingdata['color']

    def __repr__(self, ):
        return type(self).__name__

    def __hash__(self, ):
        return hash(repr(self))

    def __set_adjust(self, ) -> None:
        for column in self.raw_eatingdata.columns[1:]:
            self.raw_eatingdata[column] = self.raw_eatingdata[column] + self.adjust
        logging.info(f'adjust set: {self.adjust}')

    def __clean(self, ) -> None:
        self.raw_eatingdata = self.raw_eatingdata.loc[
            self.raw_eatingdata["Marker Name"].isin([
                "BackLeft",
                "BackRight",
                "FrontLeft",
                "FrontRight",
                "Eating",
                "EATING",
                "Grooming",
                "Approach",
            "Quiescent"
            ])
        ]

    def __match(self, ):
        self.raw_eatingdata['TimeStamp'] = funcs.get_matched_time(
                self.__tracedata.time, self.raw_eatingdata['TimeStamp'])
        self.raw_eatingdata['TimeStamp2'] = funcs.get_matched_time(
                self.__tracedata.time, self.raw_eatingdata['TimeStamp2'])

    def __set_eating_signals(self, ):
        aggregate_eating_signals = pd.DataFrame()
        for signal, _, event in self.generate_signals():
            if event == 'Interval':
                event = 'Quiescent'
            if event == "Eating":
                event = "Food Eating"
            if event == "Entry":
                event = "Food Acquisition"
            if event == "Approach":
                event = "Food Acquisition"
            signal['event'] = event
            signal['color'] = self.color_dict[event]
            aggregate_eating_signals = pd.concat(
                    [aggregate_eating_signals, signal],
                    axis=0)
        return aggregate_eating_signals.sort_index()

    def get_reorder_cols(self):
        for signal, time, approachstart, entrystart, eatingstart, eatingend in self.generate_entry_eating_signals():
            if signal.shape[0] == self.get_largest_interv():
                newsig = self.reorder(signal, time[time == entrystart].index[0], time[time == eatingend].index[0])
                return list(newsig.columns)

    @staticmethod
    def reorder(signal: pd.DataFrame, start: int, stop: int) -> pd.DataFrame:
        return signal.reindex(signal.loc[[start, stop], :].mean().sort_values(ascending=False).index, axis=1)

    def get_time_index(self, time: int | float, ):
        """Return INDEX where tracedata time matches argument num."""
        return np.where(self.__tracedata.time == time)[0][0]

    def get_signal_zscore(self, start: int | float, stop: int | float, ):
        """Return  where tracedata time matches argument num."""
        return self.__tracedata.zscores.loc[
               self.get_time_index(start):
               self.get_time_index(stop)
               ].drop(columns=['time'])

    def get_signal_time(self, start: int | float, stop: int | float, ):
        """Return  where tracedata time matches argument num."""
        return self.__tracedata.zscores['time'].loc[
               self.get_time_index(start):
               self.get_time_index(stop)]

    def get_signals_from_events(self, events: Any) -> tuple[pd.DataFrame, pd.Series]:
        signal = self.eatingdata[self.eatingdata['event'].isin(events)].drop(columns=['event'])
        color = signal.pop('color')
        return signal, color

    def get_baseline(self) -> np.ndarray:
        data = self.raw_eatingdata.to_numpy()
        baseline = data[np.where(data[:, 0] == 'Quiescent')[0][0]]
        signal = self.get_signal_zscore(baseline[1], baseline[2])
        return signal

    def get_largest_interv(self) -> int | float:
        tmp = []
        for signal, _, _, _, _, _ in self.generate_entry_eating_signals():
            tmp.append(signal.shape[0])
        return max(tmp)

    def generate_signals(self, ) -> Generator[(pd.DataFrame, str), None, None]:
        """Generator for each eating event signal (Interval(baseline), Eating, Grooming, Entry."""
        return ((
            self.get_signal_zscore(x[1], x[2]),
            self.get_signal_time(x[1], x[2]), x[0])
            for x in self.raw_eatingdata.to_numpy())

    def generate_entry_eating_signals(self, ) -> Generator[Iterable, None, None]:
        """ Generator for eating events, with entry and eating in one interval."""
        data = self.raw_eatingdata.to_numpy()
        for idx, x in (enumerate(data)):
            if idx > (len(data) - 2):
                break
            if x[0] == 'Approach'\
                    and (data[idx + 1][0] in ['BackLeft', 'BackRight', 'FrontLeft', 'FrontRight'])\
                    and data[idx + 2][0] in ['Eating', 'EATING']:
                yield (self.get_signal_zscore(x[1], data[idx + 2][2]),  # signal
                       np.round(self.get_signal_time(x[1], data[idx + 2][2]), 1),  # time
                       np.round(x[1], 2),  # approach start
                       np.round(data[idx + 1][1], 2),  # entry start
                       np.round(data[idx + 2][1], 2),  # eating start
                       np.round(data[idx + 2][2], 2))  # eating end

    def generate_eating_heatmap(
        self,
        premask: Optional[bool] = False,
        save_dir: Optional[str] = "",
        interv_size: Optional[str | float] = -np.inf,
        title: Optional[str] = '',
        **figargs
    ) -> Generator[Iterable, None, None]:
        for signal, time, approachstart, entrystart, eatingstart, eatingend in self.generate_entry_eating_signals():
            tsize = (len(signal.T.columns) / 10)
            if tsize > interv_size:
                for cell in signal:
                    signal[cell][signal[cell] < 0] = 0
                signal = signal.T
                signal.columns = np.round(np.arange(0, len(signal.columns) / 10, 0.1), 1)
                if premask:
                    premask = signal.columns
                    newcols = np.round(np.arange(
                            signal.columns[-1] + 0.1,
                            self.get_largest_interv() / 10, 0.1), 1)
                    finalsig = pd.concat([signal, pd.DataFrame(columns=newcols, index=signal.index)], axis=1)
                    data = finalsig
                else:
                    premask = None
                    data = signal
                heatmap = EatingHeatmap(
                        data,
                        premask=premask,
                        title=title,
                        save_dir=save_dir,
                        save_name=str(data.shape[1]),
                        **figargs)
                fig = heatmap.default_heatmap(eatingstart, entrystart, eatingend)
                yield fig

    def store_eating_heatmaps(
        self,
        save_dir: Optional[str] = "",
        interv_size: Optional[str | float] = -np.inf,
        title: Optional[str] = '',
        **figargs
    ) -> dict[float: EatingHeatmap]:
        heatmap_holder = {}
        for signal, time, approachstart, entrystart, eatingstart, eatingend in self.generate_entry_eating_signals():
            tsize = (len(signal.T.columns) / 10)
            cols = self.get_reorder_cols()
            if tsize > interv_size:
                signal = signal[cols]
                for cell in signal:
                    signal[cell][signal[cell] < 0] = 0
                signal = signal.T
                signal.columns = np.round(np.arange(0, len(signal.columns) / 10, 0.1), 1)
                large_cols = self.get_largest_interv() / 10
                newcols = np.round(np.arange(signal.columns[-1] + 0.1, large_cols, 0.1), 1)
                missingsig = pd.DataFrame(columns=newcols, index=signal.index)
                finalsig = pd.concat([signal, missingsig], axis=1)
                heatmap = EatingHeatmap(
                        data=finalsig,
                        title=title,
                        save_dir=save_dir,
                        **figargs)
                fig = heatmap.default_heatmap(eatingstart, entrystart, eatingend)
                heatmap_holder[tsize] = fig
        return heatmap_holder



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# setup.py

Module: Package information.
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = "1.0.0"

here = path.abspath(path.dirname(__file__))

# get the dependencies and installs
with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if "git+" not in x]
dependency_links = [
    x.strip().replace("git+", "") for x in all_reqs if x.startswith("git+")
]

setup(
    name="canalysis",
    version=__version__,
    description="Calcium Imaging trace analysis with statistics and machine learning.",
    license="MIT",
    packages=find_packages(),
    package_dir={"canalysis": "canalysis"},
    include_package_data=True,
    author="Flynn OConnell",
    zip_safe=False,
    author_email="oconnell@binghatmon.edu",
)



# -*- coding: utf-8 -*-
"""
#funcs.py

Module(util): General getter/setter/checker functions.
"""
from __future__ import annotations
import os
import logging
from pathlib import Path
from typing import Iterable, Optional, Any

import numpy as np
import pandas as pd
import scipy.stats as stats

from .wrappers import log_time
from . import excepts as e
from .wrappers import typecheck

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(message)s")


# %% COLLAPSE DATA STRUCTURES

def filter_dict(my_dict: dict, to_filter: Iterable):
    """Filter dictionary based on a list of colors."""
    return {k: v for k, v in my_dict.items() if my_dict[k] in np.unique(to_filter)}


@typecheck(str)
def check_numeric(my_str: str):
    """ Return boolean True if string is all numbers, otherwise False."""
    return my_str.isdecimal()


def check_path(my_str: str | Path):
    """ Return boolean True if string is a path, otherwise False."""
    return isinstance(my_str, Path) or any(x in my_str for x in ["/", "\\"])


def keys_exist(element, *keys):
    """Check if *keys (nested) exists in `element` (dict)"""
    if len(keys) == 0:
        raise AttributeError("keys_exists() expects at least two arguments, one given.")
    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True


@typecheck(dict, int)
def iter_events(event_dct, gap: int = 5):
    """
    Given an interval 'gap',
    iterate through a dictionary and generate an interval (start, stop) return value.
    """
    for event, ts in event_dct.items():
        intervals = interval(ts, gap)
        interv: Iterable
        for interv in intervals:
            yield event, interv


@typecheck(Iterable)
def flatten(lst: Iterable) -> list:
    return [item for sublist in lst for item in sublist]


def check_unique_path(path) -> str:
    if hasattr(path, 'stem'):
        path = path.__str__()
    filename, extension = os.path.splitext(path)
    counter = 1
    while os.path.exists(path):
        path = filename + "" + str(counter) + "" + extension
        counter += 1
    return path


@typecheck(Iterable, int)
def interval(
        lst: Iterable[any], gap: Optional[int] = 1, outer: bool = False
) -> list[tuple[Any, Any]]:
    """
    Create intervals where there elements are separated by either:
        -less than gap. 
        -more than gap.

    Args:
        lst (Iterable): Iterable to search.
        gap (int | float): length of interval.
        outer (bool): Makes larger than (gap) intervals.
    Returns:
         interv (list): New list with created interval.
    """
    interv, tmp = [], []

    for v in lst:
        if not tmp:
            tmp.append(v)
        elif abs(tmp[-1] - v) < gap:
            tmp.append(v)
        elif outer:
            interv.append(tuple((tmp[-1], v)))
            tmp = [v]
        else:
            interv.append(tuple((tmp[0], tmp[-1])))
            tmp = [v]
    return interv


@typecheck(Iterable[any], Iterable[any])
def get_peak_window(time: Iterable[any], peak: float) -> list:
    """
    Returns the index of tracedata centered 1s around the peak flourescent value for
    that trial.
    Args:
        time (list | pd.Series): List of all time values.
        peak (float) : peak time
    Returns:
         window_ind (list): list of index values to match time.
    """
    time: Iterable[any]
    aux, window_ind = [], []
    for valor in time:
        aux.append(abs(peak - valor))
    window_ind.append(aux.index(min(aux)) - 20)
    window_ind.append(aux.index(min(aux)) + 20)
    return window_ind


def get_matched_time(time: np.ndarray, match: np.ndarray) -> np.ndarray:
    """
    Finds the closest number in time to the input. Can be a single value,
    or list. Find all the absolute differences between match and time, find
    the minima of these abs. values and the argument value of the minima. Return
    all matches.
    Args:
        time : np.ndarray
            Correct values to be matched to.
        match : Iterable[any]
            Values to be matched.
    Returns
         np.ndarray of matched times.
    """
    match = np.asarray(match).reshape(-1, 1)
    mins = np.argmin(np.abs(match - time), axis=1)
    return np.array([time[mins[i]] for i in range(len(match))])
